#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements. See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# original author: Sayan Patra
"""Silverkite plotting functions."""
import sys
import warnings
from typing import Type

import numpy as np
import pandas as pd
from plotly import graph_objs as go
from plotly.subplots import make_subplots

from greykite.algo.changepoint.adalasso.changepoints_utils import get_trend_changepoint_dates_from_cols
from greykite.algo.common.model_summary import ModelSummary
from greykite.algo.forecast.silverkite.constants.silverkite_component import SilverkiteComponentsEnum
from greykite.algo.forecast.silverkite.constants.silverkite_component import SilverkiteComponentsEnumMixin
from greykite.algo.forecast.silverkite.constants.silverkite_constant import default_silverkite_constant
from greykite.common import constants as cst
from greykite.common.python_utils import get_pattern_cols
from greykite.common.viz.timeseries_plotting import add_groupby_column
from greykite.common.viz.timeseries_plotting import grouping_evaluation


class SilverkiteDiagnostics:
    """Provides various plotting functions for the model generated by the Silverkite forecast algorithms.

    Attributes
    ----------
    _silverkite_components_enum : Type[SilverkiteComponentsEnum]
        The constants for plotting the silverkite components.
    model_dict : `dict` or None
        A dict with fitted model and its attributes.
        The output of `~greykite.algo.forecast.silverkite.forecast_silverkite.SilverkiteForecast.forecast`.
    pred_category : `dict` or None
        A dictionary with keys being the predictor category and
        values being the predictors belonging to the category.
        For details, see
        `~greykite.sklearn.estimator.base_silverkite_estimator.BaseSilverkiteEstimator.pred_category`.
    time_col : str
        Name of input data time column
    value_col : str
        Name of input data value column
    components : `dict` or None
        Components of the ``SilverkiteEstimator`` model. Set by ``self.plot_components``.
        For details about the possible key values see
        `~greykite.algo.forecast.silverkite.silverkite_diagnostics.SilverkiteDiagnostics.get_silverkite_components`.
        Not available for ``random forest`` and ``gradient boosting`` methods and
        set to the default value `None`.
    model_summary : `class` or `None`
        The `~greykite.algo.common.model_summary.ModelSummary` class.
    """
    def __init__(
            self,
            constants: SilverkiteComponentsEnumMixin = default_silverkite_constant):
        self._silverkite_components_enum: Type[SilverkiteComponentsEnum] = constants.get_silverkite_components_enum()
        self.pred_category = None
        self.time_col = None
        self.value_col = None
        self.components = None
        self.model_summary = None

    def set_params(self, pred_category, time_col, value_col):
        """
        Set the various params after the model has been created.

        Parameters
        ----------
        pred_category : `dict` or None
            A dictionary with keys being the predictor category and
            values being the predictors belonging to the category.
            For details, see `~greykite.sklearn.estimator.base_silverkite_estimator.BaseSilverkiteEstimator.pred_category`.
        time_col: `str`
            Time column name in the data frame.
        value_col: `str`
            Value column name in the data frame.
        """
        self.pred_category = pred_category
        self.time_col = time_col
        self.value_col = value_col

    def summary(self, model_dict, max_colwidth=20) -> ModelSummary:
        """Creates the model summary for the given model

        Parameters
        ----------
        model_dict : `dict` or None
            A dict with fitted model and its attributes.
        max_colwidth : `int`
            The maximum length for predictors to be shown in their original name.
            If the maximum length of predictors exceeds this parameter, all
            predictors name will be suppressed and only indices are shown.

        Returns
        -------
            model_summary: `ModelSummary`
            The model summary for this model. See `~greykite.algo.common.model_summary.ModelSummary`
        """

        if model_dict is not None:
            # tree models do not have beta
            self.model_summary = ModelSummary(
                x=model_dict["x_mat"].values,
                y=model_dict["y"].values,
                pred_cols=list(model_dict["x_mat"].columns),
                pred_category=self.pred_category,
                fit_algorithm=model_dict["fit_algorithm"],
                ml_model=model_dict["ml_model"],
                max_colwidth=max_colwidth)
        else:
            self.model_summary = None
        return self.model_summary

    def plot_components(self, model_dict, names=None, title=None):
        """Class method to plot the components of a ``Silverkite`` model on the dataset passed to ``fit``.

        Parameters
        ----------
        model_dict : `dict` or None
            A dict with fitted model and its attributes.
        names: `list` [`str`], default `None`
            Names of the components to be plotted e.g. names = ["trend", "DAILY_SEASONALITY"].
            See `~greykite.algo.forecast.silverkite.silverkite_diagnostics.get_silverkite_components`
            for the full list of valid names.
            If `None`, all the available components are plotted.
        title: `str`, optional, default `None`
            Title of the plot. If `None`, default title is "Component plot".

        Returns
        -------
        fig: `plotly.graph_objs.Figure`
            Figure plotting components against appropriate time scale.
        """
        if model_dict is None:
            raise NotImplementedError("Call `self.set_params` before calling `plot_components`.")

        # recomputes `self.components` every time in case model was refit
        if not hasattr(model_dict["ml_model"], "coef_"):
            raise NotImplementedError("Component plot has only been implemented for additive linear models.")
        else:
            # Computes components for the training observations used to fit the model.
            # Observations with NAs that are dropped when fitting are not included.
            feature_df = (
                    model_dict["ml_model"].coef_
                    * model_dict["x_mat"])
            self.components = self.get_silverkite_components(
                model_dict["df_dropna"],
                self.time_col,
                self.value_col,
                feature_df)

        return self.plot_silverkite_components(
            components=self.components,
            names=names,
            title=title)

    def get_silverkite_components(
            self,
            df,
            time_col,
            value_col,
            feature_df):
        """Compute the components of a ``Silverkite`` model.

        Notes
        -----
        This function signature is chosen this way so that an user using `forecast_silverkite` can also use
        this function, without any changes to the `forecast_silverkite` function. User can compute `feature_df`
        as follows. Here `model_dict` is the output of `forecast_silverkite`.
        feature_df = model_dict["mod"].coef_ * model_dict["design_mat"]

        The function aggregates components based on the column names of `feature_df`.
        `feature_df` is defined as the patsy design matrix built by `design_mat_from_formula`
        multiplied by the corresponding coefficients, estimated by the silverkite model.

            - ``cst.TREND_REGEX``:  Used to identify `feature_df` columns corresponding to trend.
            See `greykite.common.features.timeseries_features.get_changepoint_features` for details
            about changepoint column names.
            - ``cst.SEASONALITY_REGEX``: Used to identify `feature_df` columns corresponding to seasonality.
            This means to get correct seasonalities, the user needs to provide seas_names.
            See `greykite.common.features.timeseries_features.get_fourier_col_name` for details
            about seasonality column names.
            - ``cst.EVENT_REGEX``: Used to identify `feature_df` columns corresponding to events such as holidays.
            See `~greykite.common.features.timeseries_features.add_daily_events` for details
            about event column names.

        Parameters
        ----------
        df : `pandas.DataFrame`
            A dataframe containing `time_col`, `value_col` and `regressors`.
        time_col : `str`
            The name of the time column in ``df``.
        value_col : `str`
            The name of the value column in ``df``.
        feature_df : `pandas.DataFrame`
            A dataframe containing feature columns and values.

        Returns
        -------
        components : `pandas.DataFrame`
            Contains the components of the model. Same number of rows as `df`. Possible columns are

            - `"time_col"`: same as input ``time_col``.
            - `"value_col"`: same as input ``value_col``.
            - `"trend"`: column containing the trend.
            - `"DAILY_SEASONALITY"`: column containing daily seasonality.
            - `"WEEKLY_SEASONALITY"`: column containing weekly seasonality.
            - `"MONTHLY_SEASONALITY"`: column containing monthly seasonality.
            - `"QUARTERLY_SEASONALITY"`: column containing quarterly seasonality.
            - `"YEARLY_SEASONALITY"`: column containing yearly seasonality.
            - `"events"`: column containing events e.g. holidays effect.

        """
        if feature_df is None or feature_df.empty:
            raise ValueError("feature_df must be non-empty")

        if df.shape[0] != feature_df.shape[0]:
            raise ValueError("df and feature_df must have same number of rows.")

        feature_cols = feature_df.columns
        components = df[[time_col, value_col]]

        # get trend (this includes interaction terms)
        trend_cols = get_pattern_cols(feature_cols, cst.TREND_REGEX, cst.SEASONALITY_REGEX)
        if trend_cols:
            components["trend"] = feature_df[trend_cols].sum(axis=1)

        # get seasonalities
        seas_cols = get_pattern_cols(feature_cols, cst.SEASONALITY_REGEX)
        seas_components_dict = self._silverkite_components_enum.__dict__["_member_names_"].copy()
        for seas in seas_components_dict:
            seas_pattern = self._silverkite_components_enum[seas].value.ylabel
            seas_pattern_cols = get_pattern_cols(seas_cols, seas_pattern)
            if seas_pattern_cols:
                components[seas] = feature_df[seas_pattern_cols].sum(axis=1)

        # get events (holidays for now)
        event_cols = get_pattern_cols(feature_cols, cst.EVENT_REGEX)
        if event_cols:
            components["events"] = feature_df[event_cols].sum(axis=1)

        # get trend changepoints
        # keep this column as the last column of the df
        if trend_cols:
            changepoint_dates = get_trend_changepoint_dates_from_cols(trend_cols=trend_cols)
            if changepoint_dates:
                ts = pd.to_datetime(components[time_col])
                changepoints = [1 if t in changepoint_dates else 0 for t in ts]
                components["trend_changepoints"] = changepoints

        return components

    def group_silverkite_seas_components(self, df):
        """Groups and renames``Silverkite`` seasonalities.

        Parameters
        ----------
        df: `pandas.DataFrame`
            DataFrame containing two columns:

            - ``time_col``: Timestamps of the original timeseries.
            - ``seas``: A seasonality component. It must match a component name from the
            `~greykite.algo.forecast.silverkite.constants.silverkite_component.SilverkiteComponentsEnum`.

        Returns
        -------
            `pandas.DataFrame`
            DataFrame grouped by the time feature corresponding to the seasonality
            and renamed as defined in
            `~greykite.algo.forecast.silverkite.constants.silverkite_component.SilverkiteComponentsEnum`.
        """
        time_col, seas = df.columns
        groupby_time_feature = self._silverkite_components_enum[seas].value.groupby_time_feature
        xlabel = self._silverkite_components_enum[seas].value.xlabel
        ylabel = self._silverkite_components_enum[seas].value.ylabel

        def grouping_func(grp):
            return np.nanmean(grp[seas])

        result = add_groupby_column(
            df=df,
            time_col=time_col,
            groupby_time_feature=groupby_time_feature)
        grouped_df = grouping_evaluation(
            df=result["df"],
            groupby_col=result["groupby_col"],
            grouping_func=grouping_func,
            grouping_func_name=ylabel)
        grouped_df.rename({result["groupby_col"]: xlabel}, axis=1, inplace=True)
        return grouped_df

    def plot_silverkite_components(
            self,
            components,
            names=None,
            title=None):
        """Plot the components of a ``Silverkite`` model.

        Parameters
        ----------
        components : `pandas.DataFrame`
            A dataframe containing the components of a silverkite model, similar to the output
            of `~greykite.algo.forecast.silverkite.silverkite_diagnostics.get_silverkite_components`.
        names: `list` [`str`], optional, default `None`
                Names of the components to be plotted e.g. names = ["trend", "DAILY_SEASONALITY"].
                See `~greykite.algo.forecast.silverkite.silverkite_diagnostics.get_silverkite_components`
                for the full list of valid names.
                If `None`, all the available components are plotted.
        title: `str`, optional, default `None`
                Title of the plot. If `None`, default title is "Component plot".

        Returns
        -------
        fig: `plotly.graph_objs.Figure`
            Figure plotting components against appropriate time scale.

        Notes
        -----
        If names in `None`, all the available components are plotted.
        ``value_col`` is always plotted in the first panel, as long as there is a match between
        given ``names`` list and ``components.columns``.

        See Also
        --------
        `~greykite.algo.forecast.silverkite.silverkite_diagnostics.get_silverkite_components`
        """

        time_col, value_col = components.columns[:2]
        if "trend_changepoints" in components.columns:
            trend_changepoints = components[time_col].loc[components["trend_changepoints"] == 1].tolist()
            components = components.drop("trend_changepoints", axis=1)
        else:
            trend_changepoints = None
        if names is None:
            names_kept = list(components.columns)[1:]  # do not include time_col
        else:
            # loops over components.columns to maintain the order of the components
            names_kept = [component for component in list(components.columns) if component in names]
            names_removed = set(names) - set(components.columns)

            if not names_kept:
                raise ValueError("None of the provided components have been specified in the model.")
            elif names_removed:
                warnings.warn(f"The following components have not been specified in the model: "
                              f"{names_removed}, plotting the rest.")
            if names_kept[0] != value_col:
                names_kept.insert(0, value_col)

        # Currently we use `plotly:3.10`. Need to import `v4_subplots` from `_plotly_future_`
        # to be able to use `make_subplots` function
        if 'plotly' in sys.modules:
            del sys.modules["plotly"]
            # `v4_subplots` must be imported before `plotly`
            from _plotly_future_ import v4_subplots  # isort:skip
            import plotly  # isort:skip
            assert v4_subplots  # assert statement to pass flake8 check
            assert plotly  # assert statement to pass flake8 check

        num_rows = len(names_kept)
        fig = make_subplots(rows=num_rows, cols=1, vertical_spacing=0.35 / num_rows)
        if title is None:
            title = "Component plots"
        fig.layout.update(showlegend=True, title=title, height=350 * num_rows)

        for ind, name in enumerate(names_kept):
            df = components[[time_col, name]]
            if "SEASONALITY" in name:
                df = self.group_silverkite_seas_components(df)

            xlabel, ylabel = df.columns
            row = ind + 1
            fig.append_trace(go.Scatter(
                x=df[xlabel],
                y=df[ylabel],
                name=name,
                mode="lines",
                opacity=0.8,
                showlegend=False
            ), row=row, col=1)

            # `showline = True` shows a line only along the axes. i.e. for xaxis it will line the bottom
            # of the image, but not top. Adding `mirror = True` also adds the line to the top.
            fig.update_xaxes(title_text=xlabel, showline=True, mirror=True, row=row, col=1)
            fig.update_yaxes(title_text=ylabel, showline=True, mirror=True, row=row, col=1)

        # plot trend change points
        if trend_changepoints is not None and "trend" in names_kept:
            for i, cp in enumerate(trend_changepoints):
                show_legend = (i == 0)
                fig.append_trace(
                    go.Scatter(
                        name="trend change point",
                        mode="lines",
                        x=[cp, cp],
                        y=[components["trend"].min(), components["trend"].max()],
                        line=go.scatter.Line(
                            color="#F44336",  # red 500
                            width=1.5,
                            dash="dash"),
                        showlegend=show_legend),
                    row=names_kept.index("trend") + 1,
                    col=1)

        return fig



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Reza Hosseini
"""Preprocessing function to normalize data before training."""

import warnings


def normalize_df(
        df,
        method,
        drop_degenerate_cols=True,
        replace_zero_denom=False):
    """For a given dataframe with columns including numerical values,
    it generates a function which can be applied to original data as well as
    any future data to normalize using two possible methods.
    The `"statistical"` method removes the "mean" and divides by "std".
    The `"min_max"` method removes the "minimum" and divides by the
    "maximum - minimum".
    If desired, the function also drops the columns which have only one
    possible value and can cause issues not only during normalizaton
    (returning a column with all NAs) but also potentially during fitting
    as an example.

    Parameters
    ----------
    df : `pandas.DataFrame`
        Input dataframe which (only) includes numerical values in all columns.
    method : `str`
        The method to be used for normalization.
        The "statistical" method removes the "mean" and divides by "std" for each column.
        The "min_max" method removes the "min" and divides by the "max - min"
        for each column.
    drop_degenerate_cols : `bool`, default True
        A boolean to determine if columns with only one possible value should be
        dropped in the normalized dataframe.
    replace_zero_denom : `bool`, default False
        A boolean to decide if zero denominator (e.g. standard deviation for
        ``method="statistical"``) for normalization should be replaced by 1.0.

    Returns
    -------
    normalize_info : `dict`

        A dictionary with with the main item being a normalization function.
        The items are as follows:

        ``"normalize_df_func"`` : callable (pd.DataFrame -> pd.DataFrame)
            A function which normalizes the input dataframe (``df``)
        ``"normalized_df"`` : normalized dataframe version of ``df``
        ``"keep_cols"`` : `list` [`str`]
            The list of kept columns after normalization.
        ``"drop_cols"`` : `list` [`str`]
            The list of dropped columns after normalization.
        ``"subtracted_series"`` : `pandas.Series`
            The series to be subtracted which has one value for each column of ``df``.
        ``"denominator_series"`` : `pandas.Series`
            The denominator series for normalization which has one value for each
            column of ``df``.
    """
    if method == "statistical":
        subtracted_series = df.mean()
        denominator_series = df.std()
    elif method == "min_max":
        subtracted_series = df.min()
        denominator_series = (df.max() - df.min())
    else:
        raise NotImplementedError(f"Method {method} is not implemented")

    # Replaces 0.0 in denominator series with 1.0 to avoid dividing by zero
    # when the variable has zero variance
    if replace_zero_denom:
        denominator_series.replace(to_replace=0.0, value=1.0, inplace=True)
    drop_cols = []
    keep_cols = list(df.columns)
    normalized_df = (df - subtracted_series) / denominator_series

    if drop_degenerate_cols:
        for col in df.columns:
            if normalized_df[col].isnull().any():
                drop_cols.append(col)
                warnings.warn(
                    f"{col} was dropped during normalization as it had only one "
                    "possible value (degenerate)")

        keep_cols = [col for col in list(df.columns) if col not in drop_cols]
        if len(keep_cols) == 0:
            raise ValueError(
                "All columns were degenerate (only one possible value per column).")

        subtracted_series = subtracted_series[keep_cols]
        denominator_series = denominator_series[keep_cols]

    def normalize_df_func(new_df):
        """A function which applies to a potentially new data frame (``new_df``)
        with the same columns as ``df`` (different values or row number is allowed)
        and returns a normalized dataframe with the same normalization parameters
        applied to ``df``.

        This function uses the series `subtracted_series` and
        ``denominator_series`` generated in its outer scope for normalization,
        and in this way ensures the same mapping for new data.

        Parameters
        ----------
        new_df : `pandas.DataFrame`
            Input dataframe which (only) includes numerical values in all columns.
            The columns of ``new_df`` must be the same as ``df`` which is passed to
            the outer function (``normalize_df``) to construct this function.

        Returns
        -------
        normalized_df : `pandas.dataframe`
            Normalized dataframe version of ``new_df``.
        """
        normalized_df = new_df.copy()
        if drop_degenerate_cols:
            normalized_df = normalized_df[keep_cols]
        return (normalized_df - subtracted_series) / denominator_series

    return {
        "normalize_df_func": normalize_df_func,
        "normalized_df": normalized_df,
        "keep_cols": keep_cols,
        "drop_cols": drop_cols,
        "subtracted_series": subtracted_series,
        "denominator_series": denominator_series
    }



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Kaixu Yang
"""Implements partially regularized regression."""

from typing import Union

import numpy as np
import pandas as pd
from sklearn.exceptions import ConvergenceWarning
from sklearn.linear_model import Lasso
from sklearn.linear_model import LassoCV
from sklearn.linear_model import LinearRegression
from sklearn.linear_model import Ridge
from sklearn.linear_model import RidgeCV
from sklearn.linear_model._coordinate_descent import _alpha_grid
from sklearn.model_selection import check_cv

from greykite.common.python_utils import ignore_warnings


def constant_col_finder(x, exclude_cols=None):
    """Finds constant columns in x.

    A column is considered as a constant column if all rows have the same value
    and the value is not zero.

    Parameters
    ----------
    x : `numpy.array`
        The design matrix.
    exclude_cols : `list`[`int`], default None
        Columns in ``exlucde_cols`` are not considered as constant columns.

    Returns
    -------
    constant_cols : `list`[`int`]
        A list of indices that satisfy the constant column conditions.
    """
    if exclude_cols is None:
        exclude_cols = []
    var = x.var(axis=0)
    constant_cols = [i for i in range(len(var)) if var[i] == 0 and x[0, i] != 0 and i not in exclude_cols]
    return constant_cols


class PartialRegularizeRegression:
    """Class to implement a partially regularized regression.

    A partially regularized regression is defined as

        beta_0, beta_1, beta_2 = argmin ||y - X_0beta_0 - X_1beta_1 - X_2beta_2||_2^2 + a_1 * ||beta_1||_1 + a_2 * ||beta_2||_2^2

    This regression formulation allows user to specify which regressor to have what type of regularization.
    For example, in automatic changepoint detection, "ct1" should not be penalized, yearly seasonality can have light L2 norm regularization,
    while the changepoint regressors need L1 norm regularization to be sparse.

    Attributes
    ----------
    l1_index : `list`[`int`, `str`], default None
        The index of columns that are to be penalized with L1 norm.
    l1_alpha : `float`, default 0.001
        The L1 norm regularization parameter.
    l2_index : `list`[`int`, `str`], default None
        The index of columns that are to be penalized with L2 norm.
    l2_alpha : `float`, default 0.001
        The L2 norm regularization parameter.
    has_constant : `list`[`int`]
        A list of detected constant column indices by `intercept_checker`.
    constant_index : `list`[`int`]
        A list of constant column indices after preprocessing.
    non_penalize_index : `list`[`int`]
        A list of column indices whose corresponding regressors are not to be penalized.
    intercept_ : `float`
        The fitted intercept.
    coef_ : `numpy.array`
        The fitted coefficients.
    """
    def __init__(
            self,
            l1_index=None,
            l1_alpha=0.001,
            l2_index=None,
            l2_alpha=0.001):
        """Initializes instance."""
        if l1_index is not None and l2_index is not None and len([x for x in l1_index if x in l2_index]) > 0:
            raise ValueError("l1_index and l2_index should not overlap.")
        self.l1_index = l1_index
        self.l1_alpha = l1_alpha
        self.l2_index = l2_index
        self.l2_alpha = l2_alpha
        self.has_constant = None
        self.constant_index = None
        self.non_penalize_index = None
        self.intercept_ = None
        self.intercept_ = None
        self.coef_ = None

    @ignore_warnings(category=FutureWarning)
    def fit(self, x, y):
        """Fits the partial regularized regression.

        Parameters
        ----------
        x : `numpy.array` or `pandas.DataFrame`
            The design matrix.
        y : `numpy.array`, `pandas.DataFrame` or `pandas.Series`
            The response vector.

        Returns
        -------
        self : `~greykite.algo.common.partial_regularize_regression.PartialRegularizeRegression`
            The current class instance.
        """
        # Checks if the config matches a special case.
        # Because we already checked there's no overlapping in `l1_index` and `l2_index`,
        # so we have the following 3 cases:
        #   - `l1_index` includes all columns of x: this is the same as the Lasso.
        #   - `l2_index` includes all columns of x: this is the same as the Ridge.
        #   - Both `l1_index` and `l2_index` are None: this is the same as the LinearRegression.
        if self.l1_index is not None and len(self.l1_index) == x.shape[1]:
            model = Lasso(alpha=self.l1_alpha).fit(x, y)
            self.intercept_ = model.intercept_
            self.coef_ = model.coef_
            return self
        if self.l2_index is not None and len(self.l2_index) == x.shape[1]:
            model = Ridge(alpha=self.l2_alpha).fit(x, y)
            self.intercept_ = model.intercept_
            self.coef_ = model.coef_
            return self
        if self.l1_index is None and self.l2_index is None:
            model = LinearRegression().fit(x, y)
            self.intercept_ = model.intercept_
            self.coef_ = model.coef_
            return self
        # In the above three cases, we have one type of regularization on all columns of x.
        # The cases below will be a mix of at least two regularizations.
        # We'll handle this with the core algorithm in `self._fit_analytic`.
        x, y = self._check_input(x, y)
        lasso_input = self._get_lasso_input(x, y, self.l2_alpha)
        # Fits the lasso part.
        if lasso_input["x_lasso"].shape[1] > 0:
            lasso = Lasso(alpha=self.l1_alpha, fit_intercept=False).fit(
                X=lasso_input["x_lasso"],
                y=lasso_input["y_lasso"]
            )
            beta1 = lasso.coef_
        else:
            beta1 = np.array([])
        # Computes the L2 and non-penalize part.
        beta02 = lasso_input["hl02_r"] @ (y - lasso_input["x1"] @ beta1)
        coef = np.zeros(x.shape[1])
        coef[self.non_penalize_index] = beta02[:lasso_input["x0"].shape[1]]
        coef[self.l1_index] = beta1
        coef[self.l2_index] = beta02[lasso_input["x0"].shape[1]:]
        # Extracts the intercept and coefficients.
        if not self.has_constant:
            self.intercept_ = coef[-1]
            self.coef_ = coef[:-1]
        else:
            self.intercept_ = np.sum(coef[self.constant_index])
            self.coef_ = coef
            self.coef_[self.constant_index] = 0
        return self

    def predict(self, x):
        """Predicts for new data.

        Parameters
        ----------
        x : `numpy.array` or `pandas.DataFrame`
            The new data matrix.

        Returns
        -------
        y_pred : `numpy.array`, `pandas.DataFrame` or `pandas.Series`
            The predicted values.
        """
        pred = x @ self.coef_ + self.intercept_
        return pred

    def _check_input(self, x, y):
        """Preprocesses the input design matrix, the response vector, l1_index, l2_index and constant columns.

        Parameters
        ----------
        x : `numpy.array` or `pandas.DataFrame`
            The design matrix.
        y : `numpy.array`, `pandas.DataFrame` or `pandas.Series`
            The response vector.

        Returns
        -------
        x : `numpy.array`
            The processed design matrix.
        y : `numpy.array`
            The processed response vector.
        """
        if self.l1_index is None:
            self.l1_index = []
        if self.l2_index is None:
            self.l2_index = []
        if isinstance(x, pd.DataFrame):
            columns = x.columns
        else:
            columns = []
        if not all([isinstance(x, int) for x in self.l1_index]):
            l1_index = [i for i, col in enumerate(columns) if col in self.l1_index]
            if len(l1_index) != len(self.l1_index):
                raise ValueError("l1_index takes either a list of integer indices or a list of "
                                 "a subset of the columns names in x (if x is pandas.DataFrame).")
            self.l1_index = l1_index
        if not all([isinstance(x, int) for x in self.l2_index]):
            l2_index = [i for i, col in enumerate(columns) if col in self.l2_index]
            if len(l2_index) != len(self.l2_index):
                raise ValueError("l2_index takes either a list of integer indices or a list of "
                                 "a subset of the columns names in x (if x is pandas.DataFrame).")
            self.l2_index = l2_index
        if isinstance(x, pd.DataFrame):
            x = x.values
        if isinstance(y, Union[pd.DataFrame, pd.Series].__args__):
            y = y.values
        self.has_constant = constant_col_finder(x)
        if not self.has_constant:
            x = np.concatenate([x, np.ones([x.shape[0], 1])], axis=1)
            self.constant_index = x.shape[1] - 1
        else:
            self.constant_index = self.has_constant
        self.non_penalize_index = [i for i in range(x.shape[1]) if i not in self.l1_index and i not in self.l2_index]
        return x, y

    def _get_lasso_input(self, x, y, l2_alpha):
        """Computes the transformed lasso input quantities.

        Parameters
        ----------
        x : `numpy.array`
            The design matrix.
        y : `numpy.array`
            The response vector.
        l2_alpha : `float`
            The L2 norm regularization parameter.

        Returns
        -------
        lasso_input : `dict`
            The quantites need in the lasso computation.
        """
        x0 = x[:, self.non_penalize_index]
        x1 = x[:, self.l1_index]
        x2 = x[:, self.l2_index]
        x02 = np.concatenate([x0, x2], axis=1)
        d = np.eye(x02.shape[1])
        d[:x0.shape[1], :x0.shape[1]] = np.zeros([x0.shape[1], x0.shape[1]])  # Non-penalized part is all zero.
        # This is the matrix to be multiplied to the response vector to produce the estimated beta
        # for the non-penalized and L2 norm penalized part.
        hl02_r = np.linalg.pinv(x02.T @ x02 + l2_alpha * d) @ x02.T
        hl02 = x02 @ hl02_r
        i02 = np.eye(x.shape[0])
        return {
            "x_lasso": (i02 - hl02) @ x1,
            "y_lasso": (i02 - hl02) @ y,
            "hl02": hl02,
            "hl02_r": hl02_r,
            "x0": x0,
            "x1": x1,
            "x2": x2,
            "x02": x02,
            "i02": i02
        }


class PartialRegularizeRegressionCV:
    """Implements the cross-validation version of
    `~greykite.algo.common.partial_regularize_regression.PartialRegularizeRegression`.

    Attributes
    ----------
    l1_index : `list`[`int`, `str`], default None
        The index of columns that are to be penalized with L1 norm.
    l1_alphas : `list`[`float`], default None
        The L1 norm regularization parameters.
    n_l1_alphas : `int`, default 50.
        The number of L1 norm regularization parameters. Will not be used if `l1_alphas` is provided.
    l2_index : `list`[`int`, `str`], default None
        The index of columns that are to be penalized with L2 norm.
    l2_alphas : `list`[`float`], default (0.001, 0.1, 10)
        The L2 norm regularization parameter.
    cv : `int`, cross-validation generator, or iterable, default 5
        Determines the cross-validation splitting strategy. This is the same as the ``cv`` in `sklearn` modules.
        Possible inputs for cv are:
            - None, to use the default 5-fold cross validation,
            - integer, to specify the number of folds.
            - :term:`CV splitter`,
            - An iterable yielding (train, test) splits as arrays of indices.
    has_constant : `list`[`int`]
        A list of detected constant column indices by `intercept_checker`.
    constant_index : `list`[`int`]
        A list of constant column indices after preprocessing.
    non_penalize_index : `list`[`int`]
        A list of column indices whose corresponding regressors are not to be penalized.
    best_model : `~greykite.algo.common.partial_regularize_regression.PartialRegularizeRegression`
        The best model selected by cross validation.
    best_mse : `float`
        The best cross-validated test mse.
    l1_alpha_ : `float`
        The best L1 norm regularization parameter.
    l2_alpha_ : `float`
        The best L2 norm regularization parameter.
    intercept_ : `float`
        The fitted intercept.
    coef_ : `numpy.array`
        The fitted coefficients.
    """
    def __init__(
            self,
            l1_index=None,
            l1_alphas=None,
            n_l1_alphas=50,
            l2_index=None,
            l2_alphas=(0.001, 0.1, 10),
            cv=5):
        """Initializes instance."""
        self.l1_index = l1_index
        self.l1_alphas = l1_alphas
        self.n_l1_alphas = n_l1_alphas
        self.l2_index = l2_index
        self.l2_alphas = l2_alphas
        self.cv = cv
        self.has_constant = None
        self.constant_index = None
        self.non_penalize_index = None
        self.best_model = None
        self.best_mse = None
        self.l1_alpha_ = None
        self.l2_alpha_ = None
        self.intercept_ = None
        self.coef_ = None

    @ignore_warnings(category=ConvergenceWarning)
    def fit(self, x, y):
        """Fits the partial regularized regression.

        Parameters
        ----------
        x : `numpy.array` or `pandas.DataFrame`
            The design matrix.
        y : `numpy.array`, `pandas.DataFrame` or `pandas.Series`
            The response vector.

        Returns
        -------
        self : `~greykite.algo.common.partial_regularize_regression.PartialRegularizeRegression`
            The current class instance.
        """
        if self.l1_index is not None and len(self.l1_index) == x.shape[1]:
            model = LassoCV(alphas=self.l1_alphas, n_alphas=self.n_l1_alphas).fit(x, y)
            self.intercept_ = model.intercept_
            self.coef_ = model.coef_
            self.best_model = model
            return self
        if self.l2_index is not None and len(self.l2_index) == x.shape[1]:
            model = RidgeCV(alphas=self.l2_alphas).fit(x, y)
            self.intercept_ = model.intercept_
            self.coef_ = model.coef_
            self.best_model = model
            return self
        if self.l1_index is None and self.l2_index is None:
            model = LinearRegression().fit(x, y)
            self.intercept_ = model.intercept_
            self.coef_ = model.coef_
            self.best_model = model
            return self
        x, y = self._check_input(x, y)
        cv = check_cv(self.cv)
        best_model = None
        best_mse = np.inf
        # Iterates over `l2_alphas`, because we can use `lasso_path` to iterate over `l1_alphas`.
        for l2_alpha in self.l2_alphas:
            # Finds the best `l1_alpha` and its MSE under the current `l2_alpha`.
            l1_alpha, mse = self._fit_analytic(
                x=x,
                y=y,
                l2_alpha=l2_alpha,
                cv=cv
            )
            # Updates model if improved.
            if mse < best_mse:
                best_mse = mse
                best_model = PartialRegularizeRegression(
                    l1_index=self.l1_index,
                    l1_alpha=l1_alpha,
                    l2_index=self.l2_index,
                    l2_alpha=l2_alpha,
                )
        # Fits on the entire data set with the best `l1_alpha` and `l2_alpha`.
        if not self.has_constant:
            best_model.fit(x[:, :-1], y)
        else:
            best_model.fit(x, y)
        self.best_model = best_model
        self.best_mse = best_mse
        self.l1_alpha_ = best_model.l1_alpha
        self.l2_alpha_ = best_model.l2_alpha
        self.intercept_ = best_model.intercept_
        self.coef_ = best_model.coef_
        return self

    def predict(self, x):
        """Predicts for new data.

        Parameters
        ----------
        x : `numpy.array` or `pandas.DataFrame`
            The new data matrix.

        Returns
        -------
        y_pred : `numpy.array`, `pandas.DataFrame` or `pandas.Series`
            The predicted values.
        """
        pred = self.best_model.predict(x)
        return pred

    def _check_input(self, x, y):
        """Preprocesses the input design matrix, the response vector, l1_index, l2_index and constant columns.

        Parameters
        ----------
        x : `numpy.array` or `pandas.DataFrame`
            The design matrix.
        y : `numpy.array`, `pandas.DataFrame` or `pandas.Series`
            The response vector.

        Returns
        -------
        x : `numpy.array`
            The processed design matrix.
        y : `numpy.array`
            The processed response vector.
        """
        if self.l1_index is None:
            self.l1_index = []
        if self.l2_index is None:
            self.l2_index = []
        if isinstance(x, pd.DataFrame):
            columns = x.columns
        else:
            columns = []
        if not all([isinstance(x, int) for x in self.l1_index]):
            l1_index = [i for i, col in enumerate(columns) if col in self.l1_index]
            if len(l1_index) != len(self.l1_index):
                raise ValueError("l1_index takes either a list of integer indices or a sub-list of "
                                 "columns in x (if x is pandas.DataFrame).")
            self.l1_index = l1_index
        if not all([isinstance(x, int) for x in self.l2_index]):
            l2_index = [i for i, col in enumerate(columns) if col in self.l2_index]
            if len(l2_index) != len(self.l2_index):
                raise ValueError("l2_index takes either a list of integer indices or a sub-list of "
                                 "columns in x (if x is pandas.DataFrame).")
            self.l2_index = l2_index
        if len([x for x in range(x.shape[1]) if x in self.l1_index and x in self.l2_index]) > 0:
            raise ValueError("l1_index and l2_index can not overlap.")
        if isinstance(x, pd.DataFrame):
            x = x.values
        if isinstance(y, Union[pd.DataFrame, pd.Series].__args__):
            y = y.values
        self.has_constant = constant_col_finder(x)
        if not self.has_constant:
            x = np.concatenate([x, np.ones([x.shape[0], 1])], axis=1)
            self.constant_index = x.shape[1] - 1
        else:
            self.constant_index = self.has_constant
        self.non_penalize_index = [i for i in range(x.shape[1]) if i not in self.l1_index and i not in self.l2_index]
        return x, y

    def _fit_analytic(self, x, y, l2_alpha, cv):
        """Fits the regression problem with a single L2 norm regularization parameter and
        cross validation over many L1 norm regularization parameters.

        The algorithm fits the problem with `sklearn.linear_model.LassoCV` on transformed x and y,
        then estimates the non-penalized and L2 norm penalized coefficients.

        Parameters
        ----------
        x : `numpy.array`
            The processed design matrix.
        y : `numpy.array`
            The processed response vector.
        l2_alpha : `float`
            The L2 norm regularization parameter.
        cv : `int`, cross-validation generator, or iterable, default 5
            Determines the cross-validation splitting strategy. This is the same as the ``cv`` in `sklearn` modules.
            Possible inputs for cv are:
                - None, to use the default 5-fold cross validation,
                - integer, to specify the number of folds.
                - :term:`CV splitter`,
                - An iterable yielding (train, test) splits as arrays of indices.

        Returns
        -------
        best_l1_alpha : `float`
            The best L1 regularization parameter model.
        best_mse : `float`
            The best mse.
        """
        mses = np.zeros(self.n_l1_alphas) if self.l1_index else np.zeros(1)
        if self.l1_alphas is not None:
            l1_alphas = self.l1_alphas
        else:
            l1_alphas = self._get_alphas(x, y, l2_alpha)
        for train, test in cv.split(x):
            x_train = x[train]
            x_test = x[test]
            y_train = y[train]
            y_test = y[test]
            lasso_input = self._get_lasso_input(x_train, y_train, l2_alpha)
            if l1_alphas is not None:
                # Fits all `l1_alphas`.
                path = LassoCV().path(
                    alphas=l1_alphas,
                    fit_intercept=False,
                    X=lasso_input["x_lasso"],
                    y=lasso_input["y_lasso"]
                )
                # Makes predictions and gets the best MSE.
                beta1 = path[1]
            else:
                beta1 = np.array([])
            beta02 = lasso_input["hl02_r"] @ (y_train.reshape(-1, 1) - (lasso_input["x1"] @ beta1).reshape(y_train.shape[0], -1))
            coef = np.zeros([x.shape[1], self.n_l1_alphas]) if l1_alphas is not None else np.zeros([x.shape[1], 1])
            coef[self.non_penalize_index, :] = beta02[:lasso_input["x0"].shape[1], :]
            coef[self.l1_index, :] = beta1 if l1_alphas is not None else 0
            coef[self.l2_index] = beta02[lasso_input["x0"].shape[1]:, :]
            y_pred = x_test @ coef
            mses += ((y_test.reshape(-1, 1) - y_pred) ** 2).mean(axis=0)
        best_l1_alpha = l1_alphas[mses.argmin()] if l1_alphas is not None else None
        return best_l1_alpha, np.min(mses)

    def _get_lasso_input(self, x, y, l2_alpha):
        """Computes the transformed Lasso input quantities.

        Parameters
        ----------
        x : `numpy.array`
            The design matrix.
        y : `numpy.array`
            The response vector.
        l2_alpha : `float`
            The L2 norm regularization parameter.

        Returns
        -------
        lasso_input : `dict`
            The quantites need in the Lasso computation.
        """
        x0 = x[:, self.non_penalize_index]
        x1 = x[:, self.l1_index]
        x2 = x[:, self.l2_index]
        x02 = np.concatenate([x0, x2], axis=1)
        d = np.eye(x02.shape[1])
        d[:x0.shape[1], :x0.shape[1]] = np.zeros([x0.shape[1], x0.shape[1]])
        hl02_r = np.linalg.pinv(x02.T @ x02 + l2_alpha * d) @ x02.T
        hl02 = x02 @ hl02_r
        i02 = np.eye(x.shape[0])
        return {
            "x_lasso": (i02 - hl02) @ x1,
            "y_lasso": (i02 - hl02) @ y,
            "hl02": hl02,
            "hl02_r": hl02_r,
            "x0": x0,
            "x1": x1,
            "x2": x2,
            "x02": x02,
            "i02": i02}

    def _get_alphas(self, x, y, l2_alpha):
        """Gets the Lasso alpha grid.

        Parameters
        ----------
        x : `numpy.array`
            The design matrix.
        y : `numpy.array`
            The response vector.
        l2_alpha : `float`
            The L2 norm regularization parameter.

        Returns
        -------
        l1_alpha : `list`[`float`] or None
            The alphas for L1 norm regularization to search from.
        """
        lasso_input = self._get_lasso_input(x, y, l2_alpha)
        if lasso_input["x_lasso"].shape[1] > 0:
            l1_alpha = _alpha_grid(
                lasso_input["x_lasso"],
                lasso_input["y_lasso"],
                n_alphas=self.n_l1_alphas
            )
        else:
            l1_alpha = None
        return l1_alpha



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original_author: Albert Chen
"""Constants used by `~greykite.framework."""

EVALUATION_PERIOD_CV_MAX_SPLITS = 3
"""Default value for EvaluationPeriodParam().cv_max_splits"""
COMPUTATION_N_JOBS = 1
"""Default value for ComputationParam.n_jobs"""
COMPUTATION_VERBOSE = 1
"""Default value for ComputationParam.verbose"""
CV_REPORT_METRICS_ALL = "ALL"
"""Set `cv_report_metrics` to this value to compute all metrics during CV"""
FRACTION_OUTSIDE_TOLERANCE_NAME = "OutsideTolerance"
"""Short name used to report the result of `FRACTION_OUTSIDE_TOLERANCE` in CV"""
CUSTOM_SCORE_FUNC_NAME = "score"
"""Short name used to report the result of custom `score_func` in CV"""

# UnivariateTimeSeries ``get_quantiles_and_overlays`` output column groups
MEAN_COL_GROUP = "mean"
"""Columns with mean."""
QUANTILE_COL_GROUP = "quantile"
"""Columns with quantile."""
OVERLAY_COL_GROUP = "overlay"
"""Columns with overlay."""

FORECAST_STEP_COL = "forecast_step"
"""The column name for forecast step in benchmarking"""



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Sayan Patra
"""Helper functions to framework.benchmark model templates."""

import timeit
from typing import Dict

import pandas as pd
from tqdm.autonotebook import tqdm

from greykite.common.constants import TIME_COL
from greykite.common.logging import LoggingLevelEnum
from greykite.common.logging import log_message
from greykite.framework.pipeline.pipeline import forecast_pipeline
from greykite.sklearn.cross_validation import RollingTimeSeriesSplit


def forecast_pipeline_rolling_evaluation(
        pipeline_params: Dict,
        tscv: RollingTimeSeriesSplit):
    """Runs ``forecast_pipeline`` on a rolling window basis.

    Parameters
    ----------
    pipeline_params : `Dict`
        A dictionary containing the input to the
        :func:`~greykite.framework.pipeline.pipeline.forecast_pipeline`.
    tscv : `~greykite.sklearn.cross_validation.RollingTimeSeriesSplit`
        Cross-validation object that determines the rolling window evaluation.
        See :class:`~greykite.sklearn.cross_validation.RollingTimeSeriesSplit` for details.

    Returns
    -------
    rolling_evaluation : `dict`
        Stores benchmarking results for each split, e.g.
        split_0 contains result for first split, split_1 contains result for second split and so on.
        Number of splits is determined by the input parameters.
        Every split is a dictionary with keys "runtime_sec" and "pipeline_result".
    """
    if pipeline_params["forecast_horizon"] != tscv.forecast_horizon:
        raise ValueError("Forecast horizon in 'pipeline_params' does not match that of the 'tscv'.")

    if pipeline_params["periods_between_train_test"] != tscv.periods_between_train_test:
        raise ValueError("'periods_between_train_test' in 'pipeline_params' does not match that of the 'tscv'.")

    df = pipeline_params["df"]
    time_col = pipeline_params.get("time_col", TIME_COL)
    date_format = pipeline_params.get("date_format")
    # Disables backtest. For rolling evaluation we know the actual values in forecast period.
    # So out of sample performance can be calculated using pipeline_result.forecast
    pipeline_params["test_horizon"] = 0

    rolling_evaluation = {}
    with tqdm(list(tscv.split(X=df)), ncols=800, leave=True) as progress_bar:
        for (split_num, (train, test)) in enumerate(progress_bar):
            # Description will be displayed on the left of progress bar
            progress_bar.set_description(f"Split '{split_num}' ")
            train_end_date = pd.to_datetime(
                df.iloc[train[-1]][time_col],
                format=date_format,
                infer_datetime_format=True)
            pipeline_params["train_end_date"] = train_end_date

            start_time = timeit.default_timer()
            pipeline_result = forecast_pipeline(**pipeline_params)
            runtime = timeit.default_timer() - start_time

            pipeline_output = dict(
                runtime_sec=round(runtime, 3),
                pipeline_result=pipeline_result)
            rolling_evaluation[f"split_{split_num}"] = pipeline_output

            log_message(f"Completed evaluation for split {split_num}.", LoggingLevelEnum.DEBUG)

    return rolling_evaluation



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Albert Chen
"""Represents hierarchical relationships between nodes.
Each node corresponds to a time series.
"""
import numpy as np

from greykite.common.python_utils import flatten_list


class HierarchicalRelationship:
    """Represents hierarchical relationships between nodes (time series).

    Nodes are indexed by their position in the tree, in breadth-first search (BFS) order.
    Matrix attributes such as ``bottom_up_transform`` are applied from the
    left against tree values, represented as a `numpy.array` 2D array with
    the values of each node as a row.

    Attributes
    ----------
    levels : `list` [`list` [`int]] or None
        Specifies the number of children of each parent (internal) node in the tree.
        The number of inner lists is the height of the tree. The ith inner list provides the number
        of children of each node at depth i.
        For example::

            # root node with 3 children
            levels = [[3]]

            # root node with 3 children, who have 2, 3, 3 children respectively
            levels = [[3], [2, 3, 3]]
            # These children are ordered from "left" to "right", so that the one with
            # 2 children is the first in the 2nd level.
            # This will be used as our running example.
            #           0                # level 0
            #   1       2        3       # level 1
            #  4 5    6 7 8    9 10 11   # level 2

        All leaf nodes must have the same depth. Thus, the first sublist must have one
        integer, the length of a sublist must equal the sum of the previous sublist,
        and all integers in ``levels`` must be positive.

    num_children_per_parent : `list` [`int`]
        Flattened version of ``levels``.
        The number of children for each parent (internal) node.
        [3, 2, 3, 3] in our example.
    num_internal_nodes : `int`
        The number of internal (parent) nodes (i.e. with children).
        4 in our example.
    num_leaf_nodes : `int`
        The number of leaf nodes (i.e. without children).
        8 in our example.
    num_nodes : `int`
        The total number of nodes.
        12 in our example.
    nodes_per_level : `list` [`int`]
        The number of nodes at each level of the tree.
        [1, 3, 8] in our example.
    starting_index_per_level : `list` [`int`]
        The index of the first node in each level.
        [0, 1, 4] in our example.
    starting_child_index_per_parent : `list` [`int`]
        For each parent node, the index of its first child.
        [1, 4, 6, 9]  in our example.
    sum_matrix : `numpy.array`, shape (``self.num_nodes``, ``self.num_leaf_nodes``)
        Sum matrix used to compute values of all nodes from the leaf nodes. When
        applied to a matrix with the values for leaf nodes, returns values for every
        node by bubbling up leaf node values to the internal nodes. A node's value is
        equal to the sum of its corresponding leaf nodes' values.

        ``Y_{all} = sum_matrix @ Y_{leaf}``
        In our example::

            # 4   5   6   7   8   9   10  11  (leaf nodes)
            [[1., 1., 1., 1., 1., 1., 1., 1.], # 0
             [1., 1., 0., 0., 0., 0., 0., 0.], # 1
             [0., 0., 1., 1., 1., 0., 0., 0.], # 2
             [0., 0., 0., 0., 0., 1., 1., 1.], # 3
             [1., 0., 0., 0., 0., 0., 0., 0.], # 4
             [0., 1., 0., 0., 0., 0., 0., 0.], # 5
             [0., 0., 1., 0., 0., 0., 0., 0.], # 6
             [0., 0., 0., 1., 0., 0., 0., 0.], # 7
             [0., 0., 0., 0., 1., 0., 0., 0.], # 8
             [0., 0., 0., 0., 0., 1., 0., 0.], # 9
             [0., 0., 0., 0., 0., 0., 1., 0.], # 10
             [0., 0., 0., 0., 0., 0., 0., 1.]] # 11 (all nodes)

    leaf_projection_matrix : `numpy.array`, shape (``self.num_leaf_nodes``, ``self.num_nodes``)
        Projection matrix to get leaf nodes. When applied to a matrix with the values for
        all nodes, the projection matrix selects only the rows corresponding to leaf nodes.

        ``Y_{leaf} = leaf_projection_matrix @ Y_{actual}``
        In our example::

            # 0   1   2   3   4   5   6   7   8   9   10  11  (all nodes)
            [[0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0.],  # 4
             [0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0.],  # 5
             [0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0.],  # 6
             [0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0.],  # 7
             [0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0.],  # 8
             [0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0.],  # 9
             [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0.],  # 10
             [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1.]]  # 11 (leaf nodes)

    bottom_up_transform: `numpy.array`, shape (``self.num_nodes,`` ``self.num_nodes``)
        Bottom-up transformation matrix. When applied to a matrix with the values for
        all nodes, returns values for every node by bubbling up leaf node values to the
        internal nodes. The original values of internal nodes are ignored.

        ``Y_{bu} = bottom_up_transform @ Y_{actual}``
        Note that ``bottom_up_transform = sum_matrix @ leaf_projection_matrix``.
        In our example::

            # 0   1   2   3   4   5   6   7   8   9   10  11  (all nodes)
            [[0., 0., 0., 0., 1., 1., 1., 1., 1., 1., 1., 1.], # 0
             [0., 0., 0., 0., 1., 1., 0., 0., 0., 0., 0., 0.], # 1
             [0., 0., 0., 0., 0., 0., 1., 1., 1., 0., 0., 0.], # 2
             [0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 1., 1.], # 3
             [0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0.], # 4
             [0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0.], # 5
             [0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0.], # 6
             [0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0.], # 7
             [0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0.], # 8
             [0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0.], # 9
             [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0.], # 10
             [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1.]] # 11 (all nodes)

    constraint_matrix : `numpy.array`, shape (``self.num_internal_nodes``, ``self.num_nodes``)
        Constraint matrix representing hierarchical additive constraints, where a parent's value
        is equal the sum of its leaf nodes' values.
        ``constraint_matrix @ Y_{all} = 0`` if ``Y_{all}`` satisfies the constraints.
        In our example::

            #  0    1    2    3    4    5    6    7    8    9    10   11  (all nodes)
            [[-1.,  0.,  0.,  0.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.],  # 0
             [ 0., -1.,  0.,  0.,  1.,  1.,  0.,  0.,  0.,  0.,  0.,  0.],  # 1
             [ 0.,  0., -1.,  0.,  0.,  0.,  1.,  1.,  1.,  0.,  0.,  0.],  # 2
             [ 0.,  0.,  0., -1.,  0.,  0.,  0.,  0.,  0.,  1.,  1.,  1.]]  # 3 (internal nodes)

    Methods
    -------
    get_level_of_node : callable
        Returns a node's level in the tree
    get_child_nodes : callable
        Returns the indices of a node's children in the tree
    __set_sum_matrix : callable
        Constructs the summing matrix to compute values of all nodes from the leaf nodes.
    __set_leaf_projection_matrix : callable
        Constructs leaf projection matrix to retain only values of leaf nodes.
    __set_constraint_matrix : callable
        Constructs constraint matrix that requires each parent's
        value to be the sum of its leaf node's values.
    """

    def __init__(self, levels):
        """Computes attributes given the tree structure in ``levels``.

        Parameters
        ----------
        levels : `list` [`list` [`int]] or None
            See above.
        """
        if len(levels) == 0:
            raise ValueError("`levels` must contain at least one list")

        self.levels = levels

        # Flatten the list to get the number of children for each parent
        self.num_children_per_parent = flatten_list(levels)

        # Does not allow a node to have 0 children.
        # This requirement could be relaxed in the future.
        if any([i <= 0 for i in self.num_children_per_parent]):
            raise ValueError("Every parent node must have at least one child, "
                             "so that leaf nodes are at the same depth.")

        self.num_internal_nodes = len(self.num_children_per_parent)
        # The leaf nodes are children of the last level of parents
        self.num_leaf_nodes = sum(self.levels[-1])
        self.num_nodes = self.num_internal_nodes + self.num_leaf_nodes

        # There is one root node. Sum the number of children at each level to get the number of nodes at the next level
        self.nodes_per_level = [1] + [sum(arr) for arr in levels]
        # Alternatively, directly count the number of nodes per level
        alt_nodes_per_level = [len(arr) for arr in levels] + [self.num_leaf_nodes]
        if self.nodes_per_level != alt_nodes_per_level:
            raise ValueError(
                f"The number of children does not match the expected length. Found {self.nodes_per_level}\n"
                f"Expected {alt_nodes_per_level}. The length of a sublist must equal the sum"
                f"of the numbers in the preceding sublist.")

        # Indexes nodes starting from 0 for the root node, and continuing in BFS order.
        # The first index of each level is the number of nodes in all previous levels.
        self.starting_index_per_level = [0] + list(np.cumsum(self.nodes_per_level[:-1]))  # 0-indexed, length # levels

        # For each parent node, the index of its first child relative to the first node in
        # the next level (e.g. 0th in level, 2nd in level). length self.num_internal_nodes
        within_level_child_index_per_parent = flatten_list([list(np.cumsum([0] + arr[:-1])) for arr in levels])
        # For each parent node, the index of its first child, length self.num_internal_nodes
        self.starting_child_index_per_parent = [
            # the index of the next level plus the relative offset to its first child
            self.starting_index_per_level[self.get_level_of_node(parent)+1] + within_level_child_index_per_parent[parent]
            for parent in range(self.num_internal_nodes)]

        # Sum matrix to compute all nodes from leaf nodes. Y_{all} = self.sum_matrix @ Y_{leaf}
        # Shape (self.num_nodes, self.num_leaf_nodes)
        self.sum_matrix = self.__set_sum_matrix()

        # Projection matrix to get leaf nodes. Y_{leaf} = self.leaf_projection_matrix @ Y_{all}
        # Shape (self.num_leaf_nodes, self.num_nodes)
        self.leaf_projection_matrix = self.__set_leaf_projection_matrix()

        # Transform for bottom-up transform. Y_{bu} = self.bottom_up_transform @ Y_{all}
        # Shape (self.num_nodes, self.num_nodes)
        self.bottom_up_transform = self.sum_matrix @ self.leaf_projection_matrix

        # Constraint matrix. self.constraint_matrix @ Y_{all} is 0 if Y_{all} satisfies the constraints.
        # Shape (self.num_internal_nodes, self.num_nodes)
        self.constraint_matrix = self.__set_constraint_matrix()

    def get_level_of_node(self, node):
        """Returns a node's level in the tree.
        Level is defined as the length of the path to the root.
        The root is at level 0.

        Parameters
        ----------
        node : `int`
            Index of the node.

        Returns
        -------
        level : `int`
            The level of the node in the tree.
        """
        return max([level for level, start_index in enumerate(self.starting_index_per_level) if node >= start_index])

    def get_child_nodes(self, node):
        """Returns the indices of a node's children in the tree.

        Parameters
        ----------
        node : `int`
            Index of the node.

        Returns
        -------
        child_nodes : `list` [`int`]
            Indices of all the node's children.
        """
        first_child_index = self.starting_child_index_per_parent[node]
        num_children = self.num_children_per_parent[node]
        child_nodes = list(range(first_child_index, first_child_index + num_children))
        return child_nodes

    def __set_sum_matrix(self):
        """Constructs the summing matrix.

        Returns
        -------
        sum_matrix : `numpy.array`, shape (``self.num_nodes``, ``self.num_leaf_nodes``)
            Sum matrix used to compute values of all nodes from the leaf nodes.
        """
        sum_matrix = np.zeros([self.num_nodes, self.num_leaf_nodes])

        def set_matrix_row(i, matrix):
            """Sets the final value of row ``i`` in ``matrix``.

            Consider the entry [i, j] in the matrix.
            matrix[i, j] = 1 if j is a leaf node and (i==j or j is a descendant of i),
            0 otherwise.

            Also recursively sets the values for all children of node ``i``.
            The matrix is defined bottoms-up using DFS.

            Parameters
            ----------
            i : `int`
                Node index. The function updates the row for
                this node and all its descendants.
            matrix : `numpy.array`, shape (self.num_nodes, self.num_leaf_nodes)
                Sum matrix to fill in. Passed by reference in the recursion.

            Returns
            -------
            matrix[i] : `numpy.array`
                Row i of the matrix, after filling in the proper entries.
            """
            if i >= self.num_internal_nodes:
                # leaf node is equal to itself
                leaf_node_index = i - self.num_internal_nodes
                matrix[i] = np.zeros(self.num_leaf_nodes)
                matrix[i, leaf_node_index] = 1.0
            else:
                child_nodes = self.get_child_nodes(i)
                # parent node's leaves are its children's leaves
                matrix[i] = np.sum([set_matrix_row(j, matrix) for j in child_nodes], axis=0)
            return matrix[i]

        set_matrix_row(0, sum_matrix)
        return sum_matrix

    def __set_leaf_projection_matrix(self):
        """Constructs leaf projection matrix.

        Returns
        -------
        leaf_projection_matrix : `numpy.array`, shape (``self.num_leaf_nodes``, ``self.num_nodes``)
            Projection matrix to get leaf nodes.
        """
        return np.concatenate((
            np.zeros([self.num_leaf_nodes, self.num_internal_nodes]),  # removes the parent nodes
            np.eye(self.num_leaf_nodes)  # projects the leaf nodes
        ), axis=1)

    def __set_constraint_matrix(self):
        """Constructs constraint matrix that requires each parent nodes's
        value to be the sum of its leaf nodes' values.

        Returns
        -------
        constraint_matrix : `numpy.array`, shape (``self.num_internal_nodes``, ``self.num_nodes``)
            Constraint matrix representing hierarchical additive constraints,
            where a parent's value equals the sum of its leaf nodes' values.
        """
        assert hasattr(self, "bottom_up_transform") and self.bottom_up_transform is not None
        arr = self.bottom_up_transform - np.eye(self.bottom_up_transform.shape[0])
        return arr[:self.num_internal_nodes, ]



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Reza Hosseini
"""Functions to fit a machine learning model
and use it for prediction.
"""

import random
import re
import warnings

import matplotlib
import numpy as np
import pandas as pd
import patsy
import statsmodels.api as sm
from pandas.plotting import register_matplotlib_converters
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import ElasticNetCV
from sklearn.linear_model import LarsCV
from sklearn.linear_model import LassoCV
from sklearn.linear_model import LassoLarsCV
from sklearn.linear_model import RidgeCV
from sklearn.linear_model import SGDRegressor

from greykite.algo.uncertainty.conditional.conf_interval import conf_interval
from greykite.algo.uncertainty.conditional.conf_interval import predict_ci
from greykite.common.constants import ERR_STD_COL
from greykite.common.constants import R2_null_model_score
from greykite.common.evaluation import calc_pred_err
from greykite.common.evaluation import r2_null_model_score
from greykite.common.features.normalize import normalize_df


matplotlib.use("agg")  # noqa: E402
import matplotlib.pyplot as plt  # isort:skip


register_matplotlib_converters()


def design_mat_from_formula(
        df,
        model_formula_str,
        y_col=None,
        pred_cols=None):
    """ Given a formula it extracts the response vector (y)
    and builds the design matrix (x_mat)

    :param df: pd.DataFrame
        A data frame with the response vector (y) and the feature columns (x_mat)
    :param model_formula_str: str
        A formula string e.g. "y~x1+x2+x3*x4".
        This is similar to R formulas.
        See https://patsy.readthedocs.io/en/latest/formulas.html#how-formulas-work.
    :param y_col: str
        The column name which has the value of interest to be forecasted.
        If the model_formula_str is not passed, y_col e.g. ["y"] is used
        as the response vector column
    :param pred_cols: List[str]
        The names of the feature columns
        If the model_formula_str is not passed, pred_cols e.g.
        ["x1", "x2", "x3"] is used as the design matrix columns
    :return: dict
        "y": The response vector
        "y_col": Name of the response column (y)
        "x_mat": A design matrix
        "pred_cols": Name of the columns of the design matrix (x_mat)
        "x_design_info": Information for design matrix
    """
    if model_formula_str is not None:
        y, x_mat = patsy.dmatrices(
            model_formula_str,
            data=df,
            return_type="dataframe")
        pred_cols = list(x_mat.columns)
        # get the response column name using "~" location
        y_col = re.search("(.*)~", model_formula_str).group(1).strip(" ")
        y = y[y.columns[0]]
        x_design_info = x_mat.design_info
    elif y_col is not None and pred_cols is not None:
        y = df[y_col]
        x_mat = df[pred_cols]
        x_design_info = None
    else:
        raise Exception(
            f"Either provide a model expression or both y_col and pred_cols.")

    return {
        "y": y,
        "y_col": y_col,
        "x_mat": x_mat,
        "pred_cols": pred_cols,
        "x_design_info": x_design_info}


def fit_model_via_design_matrix(
        x_train,
        y_train,
        fit_algorithm,
        sample_weight=None,
        fit_algorithm_params=None):
    """Fits the predictive model and returns the prediction function.

    Parameters
    ----------
    x_train : `numpy.array`
        The design matrix
    y_train : `numpy.array`
        The vector of responses
    fit_algorithm : `str`
        The type of predictive model used in fitting
        (implemented by sklearn and statsmodels).

        Available models are:

            - ``"statsmodels_ols"``   : `statsmodels.regression.linear_model.OLS`
            - ``"statsmodels_wls"``   : `statsmodels.regression.linear_model.WLS`
            - ``"statsmodels_gls"``   : `statsmodels.regression.linear_model.GLS`
            - ``"statsmodels_glm"``   : `statsmodels.genmod.generalized_linear_model.GLM`
            - ``"linear"``            : `statsmodels.regression.linear_model.OLS`
            - ``"elastic_net"``       : `sklearn.linear_model.ElasticNetCV`
            - ``"ridge"``             : `sklearn.linear_model.RidgeCV`
            - ``"lasso"``             : `sklearn.linear_model.LassoCV`
            - ``"sgd"``               : `sklearn.linear_model.SGDRegressor`
            - ``"lars"``              : `sklearn.linear_model.LarsCV`
            - ``"lasso_lars"``        : `sklearn.linear_model.LassoLarsCV`
            - ``"rf"``                : `sklearn.ensemble.RandomForestRegressor`
            - ``"gradient_boosting"`` : `sklearn.ensemble.GradientBoostingRegressor`

        See `~greykite.algo.common.ml_models.fit_model_via_design_matrix`
        for the sklearn and statsmodels classes that implement these methods, and their parameters.

        "linear" is the same as "statsmodels_ols", because `statsmodels.regression.linear_model.OLS`
        is more stable than `sklearn.linear_model.LinearRegression`.
    sample_weight : `numpy.array` or None, default None
        The vector of weights to be used in weighted models.
        These weights will be used to weigh each loss potentially differently.
    fit_algorithm_params : `dict` or None, default None
        Parameters passed to the requested ``fit_algorithm``.
        If None, uses the defaults defined in this function.

    Returns
    -------
    ml_model : `class`
        a trained predictive model with available predict method
    """
    fit_algorithm_dict = {
        "statsmodels_ols": sm.OLS,  # ordinary least squares
        "statsmodels_wls": sm.WLS,  # weighted least squares
        "statsmodels_gls": sm.GLS,  # generalized least squares
        "statsmodels_glm": sm.GLM,  # generalized linear models
        # "linear" has been redirected to statsmodels OLS instead of sklearn LinearRegression
        # We've found sklearn LinearRegression's solution to be unstable under some cases.
        # The reason could be that sklearn calls lapack backend that depends on the build environment.
        # statsmodels uses simple QR decomposition/pseudo inverse to compute the solution,
        # and provides stabler solutions.
        "linear": sm.OLS,
        "elastic_net": ElasticNetCV,
        "ridge": RidgeCV,
        "lasso": LassoCV,
        "sgd": SGDRegressor,  # fits linear, elastic_net, ridge, lasso via SGD. Default is ridge with alpha = 0.0001
        "lars": LarsCV,
        "lasso_lars": LassoLarsCV,
        "rf": RandomForestRegressor,
        "gradient_boosting": GradientBoostingRegressor  # only method that supports quantile loss
    }

    # for our purposes, we may want different defaults from those provided in the classes
    # sets the default `cv` and `n_estimators`
    default_fit_algorithm_params = {
        "statsmodels_ols": dict(),
        "statsmodels_wls": dict(),
        "statsmodels_gls": dict(),
        "statsmodels_glm": dict(family=sm.families.Gamma()),  # default is gamma distribution
        "linear": dict(),
        "elastic_net": dict(cv=5),
        "ridge": dict(cv=5, alphas=np.logspace(-5, 5, 100)),  # by default RidgeCV only has 3 candidate alphas, not enough
        "lasso": dict(cv=5),
        "sgd": dict(),
        "lars": dict(cv=5),
        "lasso_lars": dict(cv=5),
        "rf": dict(n_estimators=100),
        "gradient_boosting": dict()
    }

    # Re-standardizes the weights so that they sum up to data length
    # Note that in the case of no weights, each weight will be 1,
    # which also sums up to data length
    if sample_weight is not None:
        if fit_algorithm not in ["ridge", "statsmodels_wls"]:
            raise ValueError(
                "sample weights are passed. "
                f"However {fit_algorithm} does not support weighted regression.")
        sample_weight = len(sample_weight) * sample_weight / sum(sample_weight)

    if fit_algorithm not in fit_algorithm_dict.keys():
        raise ValueError(f"The fit algorithm requested was not found: {fit_algorithm}. "
                         f"Must be one of {list(fit_algorithm_dict.keys())}")

    # overwrites default params with those provided by user
    params = default_fit_algorithm_params.get(fit_algorithm, {})
    if fit_algorithm_params is not None:
        params.update(fit_algorithm_params)

    # ``ml_model`` refers to fitted machine-learning model object
    if "statsmodels" in fit_algorithm or fit_algorithm == "linear":
        if fit_algorithm == "statsmodels_wls" and sample_weight is not None:
            ml_model = fit_algorithm_dict[fit_algorithm](
                endog=y_train,
                exog=x_train,
                weights=sample_weight,
                **params)
        else:
            ml_model = fit_algorithm_dict[fit_algorithm](
                endog=y_train,
                exog=x_train,
                **params)
        ml_model = ml_model.fit()
        # Adds .coef_ and .intercept_ to statsmodels, so we could fetch parameters from .coef_ for all models.
        # Intercept is already included in params, setting .intercept_=0 in case it is needed.
        ml_model.coef_ = ml_model.params
        ml_model.intercept_ = 0.
    else:
        ml_model = fit_algorithm_dict[fit_algorithm](**params)
        if fit_algorithm == "ridge":
            ml_model.fit(
                X=x_train,
                y=y_train,
                sample_weight=sample_weight)
        else:
            ml_model.fit(
                X=x_train,
                y=y_train)

    return ml_model


def fit_ml_model(
        df,
        model_formula_str=None,
        fit_algorithm="linear",
        fit_algorithm_params=None,
        y_col=None,
        pred_cols=None,
        min_admissible_value=None,
        max_admissible_value=None,
        uncertainty_dict=None,
        normalize_method="min_max",
        regression_weight_col=None):
    """Fits predictive ML (machine learning) models to continuous
    response vector (given in ``y_col``)
    and returns fitted model.

    Parameters
    ----------
    df : pd.DataFrame
        A data frame with the response vector (y) and the feature columns
        (``x_mat``).
    model_formula_str : str
        The prediction model formula string e.g. "y~x1+x2+x3*x4".
        This is similar to R formulas.
        See https://patsy.readthedocs.io/en/latest/formulas.html#how-formulas-work.
    fit_algorithm : `str`, optional, default "linear"
        The type of predictive model used in fitting.

        See `~greykite.algo.common.ml_models.fit_model_via_design_matrix`
        for available options and their parameters.
    fit_algorithm_params : `dict` or None, optional, default None
        Parameters passed to the requested fit_algorithm.
        If None, uses the defaults in `~greykite.algo.common.ml_models.fit_model_via_design_matrix`.
    y_col : str
        The column name which has the value of interest to be forecasted
        If the model_formula_str is not passed, ``y_col`` e.g. ["y"]
        is used as the response vector column
    pred_cols : List[str]
        The names of the feature columns
        If the ``model_formula_str`` is not passed, ``pred_cols`` e.g.
        ["x1", "x2", "x3"] is used as the design matrix columns
    min_admissible_value : Optional[Union[int, float, double]]
        the minimum admissible value for the ``predict`` function to return
    max_admissible_value : Optional[Union[int, float, double]]
        the maximum admissible value for the ``predict`` function to return
    uncertainty_dict : `dict` or None
        If passed as a dictionary an uncertainty model will be fit.
        The items in the dictionary are:

            ``"uncertainty_method"`` : `str`
                the title of the method
                as of now only "simple_conditional_residuals" is implemented
                which calculates CIs by using residuals
            ``"params"`` : `dict`
                A dictionary of parameters needed for the ``uncertainty_method``
                requested

    normalize_method : `str` or None, default "min_max"
        If a string is provided, it will be used as the normalization method
        in `~greykite.common.features.normalize.normalize_df`, passed via
        the argument ``method``. Available options are: "min_max", "statistical".
        If None, no normalization will be performed.
        See that function for more details.
    regression_weight_col : `str` or None, default None
        The column name for the weights to be used in weighted regression version
        of applicable machine-learning models.

    Returns
    -------
    trained_model : `dict`
        Trained model dictionary with keys:

            ``"ml_model"`` : A trained model with predict method
            ``"uncertainty_model"`` : `dict`
                The returned uncertainty_model dict from
                `~greykite.algo.uncertainty.conditional.conf_interval.conf_interval`.
    """

    # build model matrices
    res = design_mat_from_formula(
        df=df,
        model_formula_str=model_formula_str,
        y_col=y_col,
        pred_cols=pred_cols)

    y = res["y"]
    x_mat = res["x_mat"]
    y_col = res["y_col"]
    x_design_info = res["x_design_info"]

    normalize_df_func = None
    if normalize_method is not None:
        if "Intercept" in (x_mat.columns):
            cols = [col for col in list(x_mat.columns) if col != "Intercept"]
        else:
            cols = list(x_mat.columns)
        normalize_info = normalize_df(
            df=x_mat[cols],
            method=normalize_method,
            drop_degenerate_cols=False,
            replace_zero_denom=True)
        x_mat[cols] = normalize_info["normalized_df"]
        x_mat = x_mat.fillna(value=0)
        normalize_df_func = normalize_info["normalize_df_func"]

    sample_weight = None
    if regression_weight_col is not None:
        if df[regression_weight_col].min() < 0:
            raise ValueError(
                "Weights can not be negative. "
                f"The column {regression_weight_col} includes negative values.")
        sample_weight = df[regression_weight_col]

    # prediction model generated by using all observed data
    ml_model = fit_model_via_design_matrix(
        x_train=x_mat,
        y_train=y,
        fit_algorithm=fit_algorithm,
        fit_algorithm_params=fit_algorithm_params,
        sample_weight=sample_weight)

    # uncertainty model is fitted if uncertainty_dict is passed
    uncertainty_model = None
    if uncertainty_dict is not None:
        uncertainty_method = uncertainty_dict["uncertainty_method"]
        if uncertainty_method == "simple_conditional_residuals":
            # reset index to match behavior of predict before assignment
            new_df = df.reset_index(drop=True)
            (new_x_mat,) = patsy.build_design_matrices(
                [x_design_info],
                data=new_df,
                return_type="dataframe")
            if normalize_df_func is not None:
                if "Intercept" in list(x_mat.columns):
                    cols = [col for col in list(x_mat.columns) if col != "Intercept"]
                else:
                    cols = list(x_mat.columns)
                new_x_mat[cols] = normalize_df_func(new_x_mat[cols])
            new_x_mat = new_x_mat.fillna(value=0)
            new_df[f"{y_col}_pred"] = ml_model.predict(new_x_mat)
            new_df["fit_residual"] = new_df[y_col] - new_df[f"{y_col}_pred"]

            # re-assign some param defaults for function conf_interval
            # with values best suited to this case
            conf_interval_params = {
                "quantiles": [0.025, 0.975],
                "sample_size_thresh": 10}

            if uncertainty_dict["params"] is not None:
                conf_interval_params.update(uncertainty_dict["params"])
            uncertainty_model = conf_interval(
                df=new_df,
                value_col=y_col,
                residual_col="fit_residual",
                min_admissible_value=min_admissible_value,
                max_admissible_value=max_admissible_value,
                **conf_interval_params)
        else:
            raise NotImplementedError(
                f"uncertainty method: {uncertainty_method} is not implemented")

    # We get the model summary for a subset of models
    # where summary is available (statsmodels module),
    # or summary can be constructed (a subset of models from sklearn).
    ml_model_summary = None
    if "statsmodels" in fit_algorithm:
        ml_model_summary = ml_model.summary()
    elif hasattr(ml_model, "coef_"):
        var_names = list(x_mat.columns)
        coefs = ml_model.coef_
        ml_model_summary = pd.DataFrame({
            "variable": var_names,
            "coef": coefs})

    trained_model = {
        "x_design_info": x_design_info,
        "ml_model": ml_model,
        "uncertainty_model": uncertainty_model,
        "ml_model_summary": ml_model_summary,
        "y_col": y_col,
        "x_mat": x_mat,
        "min_admissible_value": min_admissible_value,
        "max_admissible_value": max_admissible_value,
        "normalize_df_func": normalize_df_func,
        "regression_weight_col": regression_weight_col}

    if uncertainty_dict is None:
        fitted_df = predict_ml(
            fut_df=df,
            trained_model=trained_model)
    else:
        fitted_df = predict_ml_with_uncertainty(
            fut_df=df,
            trained_model=trained_model)

    trained_model["fitted_df"] = fitted_df

    return trained_model


def predict_ml(
        fut_df,
        trained_model):
    """Returns predictions on new data using the machine-learning (ml) model
    fitted via ``fit_ml_model``.

    :param fut_df: pd.DataFrame
        Input data for prediction.
        Must have all columns used for training,
        specified in ``model_formula_str`` or ``pred_cols``
    :param trained_model: dict
        A trained model returned from ``fit_ml_model``
    :return: pd.DataFrame
        Input data with ``y_col`` set to the predicted values
    """
    y_col = trained_model["y_col"]
    ml_model = trained_model["ml_model"]
    x_design_info = trained_model["x_design_info"]
    min_admissible_value = trained_model["min_admissible_value"]
    max_admissible_value = trained_model["max_admissible_value"]

    # reset indices to avoid issues when adding new cols
    fut_df = fut_df.reset_index(drop=True)
    (x_mat,) = patsy.build_design_matrices(
        [x_design_info],
        data=fut_df,
        return_type="dataframe")
    if trained_model["normalize_df_func"] is not None:
        if "Intercept" in list(x_mat.columns):
            cols = [col for col in list(x_mat.columns) if col != "Intercept"]
        else:
            cols = list(x_mat.columns)
        x_mat[cols] = trained_model["normalize_df_func"](x_mat[cols])
    x_mat = x_mat.fillna(value=0)
    y_pred = ml_model.predict(x_mat)
    if min_admissible_value is not None or max_admissible_value is not None:
        y_pred = np.clip(
            a=y_pred,
            a_min=min_admissible_value,
            a_max=max_admissible_value)
    fut_df[y_col] = y_pred.tolist()
    return fut_df


def predict_ml_with_uncertainty(
        fut_df,
        trained_model):
    """Returns predictions and prediction intervals on new data using
    the machine-learning (ml) model
    fitted via ``fit_ml_model`` and the uncertainty model fitted via
    ``greykite.algo.uncertainty.conditional.conf_interval.conf_interval``

    :param fut_df: pd.DataFrame
        Input data for prediction.
        Must have all columns specified by
        ``model_formula_str`` or ``pred_cols``
    :param trained_model: dict
        A trained model returned from ``fit_ml_model``
    :return: pd.DataFrame
        Input data with ``y_col`` set to the predicted values
        and ``f"{y_col}_quantile_summary"`` set to the uncertainty
    """
    # gets point predictions
    fut_df = fut_df.reset_index(drop=True)
    y_col = trained_model["y_col"]
    y_pred = predict_ml(
        fut_df=fut_df,
        trained_model=trained_model)[y_col]
    fut_df[y_col] = y_pred.tolist()

    # apply uncertainty model
    pred_df_with_uncertainty = predict_ci(
        fut_df,
        trained_model["uncertainty_model"])
    # add uncertainty column to df
    pred_df_with_uncertainty.reset_index(drop=True, inplace=True)
    fut_df.reset_index(drop=True, inplace=True)
    fut_df[f"{y_col}_quantile_summary"] = (
        pred_df_with_uncertainty[f"{y_col}_quantile_summary"])
    fut_df[ERR_STD_COL] = pred_df_with_uncertainty[ERR_STD_COL]
    return fut_df


def fit_ml_model_with_evaluation(
        df,
        model_formula_str=None,
        y_col=None,
        pred_cols=None,
        fit_algorithm="linear",
        fit_algorithm_params=None,
        ind_train=None,
        ind_test=None,
        training_fraction=0.9,
        randomize_training=False,
        min_admissible_value=None,
        max_admissible_value=None,
        uncertainty_dict=None,
        normalize_method="min_max",
        regression_weight_col=None):
    """Fits prediction models to continuous response vector (y)
    and report results.

    Parameters
    ----------
    df : `pandas.DataFrame`
        A data frame with the response vector (y) and the feature columns (``x_mat``)
    model_formula_str : `str`
        The prediction model formula e.g. "y~x1+x2+x3*x4".
        This is similar to R language (https://www.r-project.org/) formulas.
        See https://patsy.readthedocs.io/en/latest/formulas.html#how-formulas-work.
    y_col : `str`
        The column name which has the value of interest to be forecasted
        If the ``model_formula_str`` is not passed, ``y_col`` e.g. ["y"] is used as the response
        vector column
    pred_cols : `list` [`str`]
        The names of the feature columns
        If the ``model_formula_str`` is not passed, ``pred_cols`` e.g. ["x1", "x2", "x3"]
        is used as the design matrix columns
    fit_algorithm : `str`, optional, default "linear"
        The type of predictive model used in fitting.

        See `~greykite.algo.common.ml_models.fit_model_via_design_matrix`
        for available options and their parameters.
    fit_algorithm_params : `dict` or None, optional, default None
        Parameters passed to the requested fit_algorithm.
        If None, uses the defaults in `~greykite.algo.common.ml_models.fit_model_via_design_matrix`.
    ind_train : `list` [`int`]
        The index (row number) of the training set
    ind_test : `list` [`int`]
        The index (row number) of the test set
    training_fraction : `float`, between 0.0 and 1.0
        The fraction of data used for training
        This is invoked if ind_train and ind_test are not passed
        If this is also None or 1.0, then we skip testing
        and train on the entire dataset
    randomize_training : `bool`
        If True, then the training and the test sets will be randomized
        rather than in chronological order
    min_admissible_value : Optional[Union[int, float, double]]
        The minimum admissible value for the ``predict`` function to return
    max_admissible_value : Optional[Union[int, float, double]]
        The maximum admissible value for the ``predict`` function to return
    uncertainty_dict: `dict` or None
        If passed as a dictionary an uncertainty model will be fit.
        The items in the dictionary are:

            ``"uncertainty_method"`` : `str`
                the title of the method
                as of now only "simple_conditional_residuals" is implemented
                which calculates CIs by using residuals
            ``"params"`` : `dict`
                A dictionary of parameters needed for the ``uncertainty_method``
                requested

    normalize_method : `str` or None, default "min_max"
        If a string is provided, it will be used as the normalization method
        in `~greykite.common.features.normalize.normalize_df`, passed via
        the argument ``method``. Available options are: "min_max", "statistical".
        If None, no normalization will be performed.
        See that function for more details.
    regression_weight_col : `str` or None, default None
        The column name for the weights to be used in weighted regression version
        of applicable machine-learning models.

    Returns
    -------
    trained_model : `dict`
        Trained model dictionary with the following keys.

            "ml_model": A trained model object
            "summary": Summary of the final model trained on all data
            "x_mat": Feature vectors matrix used for training of full data (rows of ``df`` with NA are dropped)
            "y": Response vector for training and testing (rows of ``df`` with NA are dropped).
                The index corresponds to selected rows in the input ``df``.
            "y_train": Response vector used for training
            "y_train_pred": Predicted values of ``y_train``
            "training_evaluation": score function value of ``y_train`` and ``y_train_pred``
            "y_test": Response vector used for testing
            "y_test_pred": Predicted values of ``y_test``
            "test_evaluation": score function value of ``y_test`` and ``y_test_pred``
            "uncertainty_model": `dict`
                The returned uncertainty_model dict from
                `~greykite.algo.uncertainty.conditional.conf_interval.conf_interval`.
            "plt_compare_test": plot function to compare ``y_test`` and ``y_test_pred``,
            "plt_pred": plot function to compare
                ``y_train``, ``y_train_pred``, ``y_test`` and ``y_test_pred``.

    """
    # to avoid pandas unnecessary warnings due to chain assignment
    pd.options.mode.chained_assignment = None

    # dropping NAs
    if df.isnull().values.any():
        nrows_original = df.shape[0]
        df = df.dropna(subset=df.columns)  # preserves index
        nrows = df.shape[0]
        warnings.warn(
            f"The data frame included {nrows_original-nrows} row(s) with NAs which were removed for model fitting.",
            UserWarning)
        if nrows <= 2:
            raise ValueError(
                f"Model training requires at least 3 observations, but the dataframe passed "
                f"to training has {nrows} rows after removing NAs."
                f"Sometimes this can be caused by unnecessary columns in your training data "
                f"which contain NAs. Make sure to remove unnecessary columns from data before "
                f"passing it to the function.")

    # an internal function for fitting model
    # this is wrapped into a function since we can do evaluations
    trained_model = fit_ml_model(
        df=df,
        model_formula_str=model_formula_str,
        fit_algorithm=fit_algorithm,
        fit_algorithm_params=fit_algorithm_params,
        y_col=y_col,
        pred_cols=pred_cols,
        min_admissible_value=min_admissible_value,
        max_admissible_value=max_admissible_value,
        uncertainty_dict=uncertainty_dict,
        normalize_method=normalize_method,
        regression_weight_col=regression_weight_col)

    # we store the obtained ``y_col`` from the function in a new variable (``y_col_final``)
    # this is done since the input y_col could be None
    # in which case we extract ``y_col`` from the formula (``model_formula_str``)
    y_col_final = trained_model["y_col"]

    # determining what should be the training and test sets
    n = len(df)
    skip_test = False
    if ind_train is not None and ind_test is not None:
        if max(ind_train) >= min(ind_test):
            raise Exception("Test set indices should start after training set indices.")
        elif max(ind_test) >= n:
            warnings.warn(
                "Testing set indices exceed the size of the dataset."
                "Setting max index of the Test set "
                "equal to the max index of the dataset.")
            ind_test = [x for x in ind_test if x < n]
    elif ind_train is None and training_fraction is not None and training_fraction < 1.0:
        k = round(n * training_fraction)
        k = int(k)
        ind_train = range(k)
        if randomize_training:
            ind_train = random.sample(range(n), k)
            ind_train = np.sort(ind_train)
        ind_test = list(set(range(n)) - set(ind_train))
        ind_test = np.sort(ind_test)
    else:
        ind_train = range(n)
        ind_test = []
        skip_test = True

    df_train = df.iloc[ind_train]
    df_test = df.iloc[ind_test]
    y_train = df_train[y_col_final]
    y_test = df_test[y_col_final]

    if skip_test:
        y_test_pred = None
        test_evaluation = None
        plt_compare_test = None
        y_train_pred = predict_ml(
            fut_df=df,
            trained_model=trained_model,
            )[y_col_final].tolist()

        def plt_pred():
            plt.plot(ind_train, y_train, label="full data", alpha=0.4)
            plt.plot(ind_train, y_train_pred, label="fit", alpha=0.4)
            plt.xlabel("index")
            plt.ylabel("value")
            plt.title("fit on the whole dataset")
            plt.legend()
    else:
        # validation: fit with df_train only and predict with df_test
        # first remove responses from df_test
        df_test[y_col_final] = None
        trained_model_tr = fit_ml_model(
            df=df_train,
            model_formula_str=model_formula_str,
            fit_algorithm=fit_algorithm,
            y_col=y_col,
            pred_cols=pred_cols,
            min_admissible_value=min_admissible_value,
            max_admissible_value=max_admissible_value,
            uncertainty_dict=uncertainty_dict,
            normalize_method=normalize_method,
            regression_weight_col=regression_weight_col)

        y_train_pred = predict_ml(
            fut_df=df_train,
            trained_model=trained_model_tr)[y_col_final].tolist()

        y_test_pred = predict_ml(
            fut_df=df_test,
            trained_model=trained_model_tr)[y_col_final].tolist()

        test_evaluation = calc_pred_err(
            y_true=y_test,
            y_pred=y_test_pred)

        test_evaluation[R2_null_model_score] = r2_null_model_score(
            y_true=y_test,
            y_pred=y_test_pred,
            y_pred_null=y_train.mean(),
            y_train=None)

        def plt_compare_test():
            plt.scatter(y_test, y_test_pred, color="red", alpha=0.05)
            plt.xlabel("observed")
            plt.ylabel("predicted")
            plt.title("test set")

        def plt_pred():
            plt.plot(ind_train, y_train, label="train", alpha=0.4)
            plt.plot(ind_train, y_train_pred, label="fit", alpha=0.4)
            plt.plot(ind_test, y_test, label="observed test set", alpha=0.4)
            plt.plot(ind_test, y_test_pred, label="predicted test set", alpha=0.4)
            plt.xlabel("index")
            plt.ylabel("value")
            plt.title("training and test fits")
            plt.legend()

    training_evaluation = calc_pred_err(
        y_true=y_train,
        y_pred=y_train_pred)
    training_evaluation[R2_null_model_score] = r2_null_model_score(
        y_true=y_train,
        y_pred=y_train_pred,
        y_pred_null=y_train.mean(),
        y_train=None)

    trained_model["summary"] = None
    trained_model["y"] = df[y_col_final]
    trained_model["y_train"] = y_train
    trained_model["y_train_pred"] = y_train_pred
    trained_model["training_evaluation"] = training_evaluation
    trained_model["y_test"] = y_test
    trained_model["y_test_pred"] = y_test_pred
    trained_model["test_evaluation"] = test_evaluation
    trained_model["plt_compare_test"] = plt_compare_test
    trained_model["plt_pred"] = plt_pred

    return trained_model



# BSD 2-CLAUSE LICENSE

# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:

# Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
# #ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# original author: Rachit Arora

from typing import Type


class SilverkiteHoliday:
    """Holiday constants to be used by Silverkite"""

    HOLIDAY_LOOKUP_COUNTRIES_AUTO = ("UnitedStates", "UnitedKingdom", "India", "France", "China")
    """Auto setting for the countries that contain the holidays to include in the model"""

    HOLIDAYS_TO_MODEL_SEPARATELY_AUTO = (
        "New Year's Day",
        "Chinese New Year",
        "Christmas Day",
        "Independence Day",
        "Thanksgiving",
        "Labor Day",
        "Good Friday",
        "Easter Monday [England, Wales, Northern Ireland]",
        "Memorial Day",
        "Veterans Day")
    """Auto setting for the holidays to include in the model"""

    ALL_HOLIDAYS_IN_COUNTRIES = "ALL_HOLIDAYS_IN_COUNTRIES"
    """Value for `holidays_to_model_separately` to request all holidays in the lookup countries"""

    HOLIDAYS_TO_INTERACT = (
        "Christmas Day",
        "Christmas Day_minus_1",
        "Christmas Day_minus_2",
        "Christmas Day_plus_1",
        "Christmas Day_plus_2",
        "New Years Day",
        "New Years Day_minus_1",
        "New Years Day_minus_2",
        "New Years Day_plus_1",
        "New Years Day_plus_2",
        "Thanksgiving",  # always on a Thursday, so Thursday/Friday may be affected
        "Thanksgiving_plus_1",
        "Independence Day")
    """Significant holidays that may have a different daily seasonality pattern"""


class SilverkiteHolidayMixin:
    """Provides a mixin for the SilverkiteHoliday constants"""

    def get_silverkite_holiday(self) -> Type[SilverkiteHoliday]:
        """Return the SilverkiteHoliday constants"""
        return SilverkiteHoliday
